function out = general_solver(args)
% Approximate the solution to 
%
% \dot(P) = A'P + PA + Q - PB Rinv B'P  , P(0) = P0,
%
% by taking Nt time steps of size h with the splitting scheme specified in
% args.solver.
%
% Here,
% P0 = L0 D0 L0' and Q = IQL IQD IQL'
%
% All inputs submitted in argument structure args. 
%
% Outputs: t, time steps
%          Ls, low-rank factors L at time steps (if args.intermediates)
%          otherwise only at final time
%          Ds, as above, but for D in LDL^T
%          ms, as above, but for the ranks of L
%          time, estimated total elapsed computation time, excluding
%                unpacking and trivial setup
%          timeIQ, estimated computation time for evaluating the
%                  integral IQ(h/2)

%
% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.
%

    % Unpack most of args for clarity
    B = args.B; Rinv = args.Rinv;  M = args.M;
    LQ = args.LQ; DQ = args.DQ; L0 = args.L0; D0 = args.D0;    
    t0 = args.t0;  tend = args.tend;  Nt = args.Nt;
    intermediates = args.intermediates; print_step = args.print_step; 
    
    % Check for time dependency of A, B, C, M
    if ~isfield(args.impODE,'method') || (isfield(args.impODE,'method') && strcmp(args.impODE.method, 'timevarying'))
        timevarying = true;
        args.timevarying = true;
    else
        timevarying = false;
        args.timevarying = false;
    end
    
    if ~isfield(args, 'quadrature') || ~isfield(args.quadrature, 'order')
        if strcmp(args.solver, 'Lie')
            args.quadrature.order = 3;
        elseif strcmp(args.solver, 'Strang')
            args.quadrature.order = 5;
        elseif strcmp(args.solver, 'additive_symmetric')
            args.quadrature.order = args.order + 3;
        elseif strcmp(args.solver, 'additive_asymmetric')
            s = args.order;
            if mod(s, 2) == 0 % even
                args.quadrature.order = s + 3;
            else % odd
                args.quadrature.order = s + 2;
            end
        end
    end
    
    if ~isfield(args, 'save_solution')
        save_solution = true;
    else
        save_solution = args.save_solution;
    end
    
    h = (tend-t0) / Nt;
    t = linspace(t0, tend, Nt+1);    
        
    ms = zeros(Nt, 1);
    ms(1) = size(L0,2);
    
    if intermediates
        if save_solution
            out.Ls{Nt+1} = [];
            out.Ls{1} = L0;
            out.Ds{Nt+1} = [];
            out.Ds{1} = D0;
        end
        out.Ks{Nt+1} = [];
        if ~timevarying
            out.Ks{1} = Rinv*(B'*L0)*D0*(L0'*M);
        else
            out.Ks{1} = Rinv*(B(t0)'*L0)*D0*(L0'*M(t0));
        end
    else
        if save_solution
            out.Ls = L0;
            out.Ds = D0;
        end
        if ~timevarying
            out.Ks = Rinv*(B'*L0)*D0*(L0'*M);
        else
            out.Ks = Rinv*(B(t0)'*L0)*D0*(L0'*M(t0));
        end
    end
    L = L0;
    D = D0;
    
    args.impODEobj = impODEobj(args, true);

    if strcmp(args.solver, 'Lie')
        hfactors = [1];
        step_function = @Lie_step;
        name = 'Lie';
    elseif strcmp(args.solver, 'Strang')
        hfactors = [1/2];
        step_function = @Strang_step;
        name = 'Strang';
    elseif strcmp(args.solver, 'Symmetric')
        s = args.order / 2;
        hfactors = 1./(1:s);
        name = ['Symmetric, order ', int2str(2*s)];
        % Method coefficients
        switch 2*s
            case 2
                gamma = [1/2];
            case 4
                gamma = [-1/6, 2/3];
            case 6
                gamma = [1/48; -8/15; 81/80];
            case 8
                gamma = [-1/720; 8/45; -729/560; 512/315];
            otherwise
                gamma = compute_additive_coefficients(2*s, true);
        end
        step_function = @(h, B, IQL, IQD, Rinv, L, D, args, t) ...
                         additive_symmetric_step(s, gamma, h, B, IQL, IQD, L, D, Rinv, args, t);

    elseif strcmp(args.solver, 'Asymmetric')
        s = args.order;
        hfactors = 1./(1:s);
        name = ['Asymmetric, order ', int2str(s)];
            % Method coefficients
        switch s
            case 1
                gamma = [1];
            case 2
                gamma = [-1, 2];
            case 3
                gamma = [1/2, -4, 9/2];
            case 4
                gamma = [-1/6, 4, -27/2, 32/3];
            otherwise
                gamma = compute_additive_coefficients(s, false);
        end
        step_function = @(h, B, IQL, IQD, Rinv, L, D, args, t) ...
                         additive_asymmetric_step(s, gamma, h, B, IQL, IQD, L, D, Rinv, args, t);

    end
    
    
    telapsed = tic;

    % If not time-varying, we may approximate the integral once and for all
    if ~timevarying
        if args.print_output
            fprintf('Computing low-rank factorization of integral I_Q. \n');
        end
        timeIQ = tic;
        [IQL, IQD] = compute_integral_factors_LDLT(LQ, DQ, h, hfactors, args);
        IQL = IQL{1};
        IQD = IQD{1};
        timeIQ = toc(timeIQ);
    else
        timeIQ = 0;
    end
   
    if args.print_output
        fprintf('Starting time-stepping. \n');
    end
    
    for j = 1:Nt
        if print_step
            fprintf(['Step: %d of %d. (', name, ') \n'], j, Nt);
        end
        
        % If time-varying, we have to compute a new integral term in each
        % step.
        if timevarying
            tIQ = tic;
            [IQL, IQD] = compute_integral_factors_LDLT(LQ, DQ, h, hfactors, args, [t(j), t(j) + h/2]);
            timeIQ = timeIQ + toc(tIQ);
        end
        
        if intermediates
            [L, D] = step_function(h, B, IQL, IQD, Rinv, L, D, args, t(j));
            if save_solution
                out.Ls{j+1} = L;
                out.Ds{j+1} = D;
            end
            out.ms(j+1) = size(L,2);
            if ~timevarying
                out.Ks{j+1} = Rinv*(B'*L)*D*(L'*M);
            else
                out.Ks{j+1} = Rinv*(B(t(j+1))'*L)*D*(L'*M(t(j+1)));
            end

        else
            [L, D] = step_function(h, B, IQL, IQD, Rinv, L, D, args, t(j));
            if save_solution
                out.Ls = L;
                out.Ds = D;
            end
            out.ms(j+1) = size(L,2);
            if ~timevarying
                out.Ks = Rinv*(B'*L)*D*(L'*M);
            else
                out.Ks = Rinv*(B(t(j+1))'*L)*D*(L'*M(t(j+1)));
            end
        end
        
        if args.print_output
            fprintf('Size of solution low-rank factor: %d \n', out.ms(j+1));
        end
        
    end
    time = toc(telapsed);
    out.timeIQ = timeIQ;
    out.time = time;
    
end